"""The pytest fixtures
   https://docs.pytest.org/en/latest/fixture.html
"""
# pylint: disable=missing-function-docstring, unused-import, wildcard-import, unused-wildcard-import, line-too-long, unused-argument
import pytest
from icpp.conftest_base import *  # pytest fixtures provided by icpp

# Define your pytest fixtures below



"""Handles 'icpp init' """

import sys
import shutil
import typer
from icpp.__main__ import app
from icpp import config_default


@app.command()
def init() -> None:
    """Copies the greet project into a new subdirectory.

    See the Getting Started tutorial: https://docs.icpp.world/getting-started.html
    """
    icpp_init_path = config_default.ICPP_INIT_PATH
    local_init_path = config_default.LOCAL_INIT_PATH

    typer.echo("--")
    if local_init_path.exists():
        typer.echo(
            f"ERROR: Directory '{local_init_path}' already exists.\n"
            f"       Please remove it first."
        )
        sys.exit(1)

    shutil.copytree(icpp_init_path, local_init_path)
    typer.echo("Successfully created the `greet` project.")



"""Reads the icpp.toml file"""

import sys
import glob
from pathlib import Path
from typing import Any, Dict, List, Set
import typer

# `tomllib` was introduced in python 3.11
# for earlier versions, use `tomli` instead
if sys.version_info >= (3, 11):
    import tomllib
else:
    try:
        import tomli as tomllib
    except ImportError:
        typer.echo("ERROR: cannot find python module tomli")
        sys.exit(1)

# Look for `icpp.toml`in the current folder
icpp_toml_path = Path("icpp.toml")
if not icpp_toml_path.exists():
    typer.echo(
        f"ERROR: There is no `icpp.toml` file in the current working directory: \n"
        f"        {icpp_toml_path.parent.resolve()} \n"
        f"       Maybe you're in the wrong folder, \n"
        f"       or else can you please create one, so I know what to build?"
    )
    sys.exit(1)


def validate(d_in: Dict[Any, Any]) -> None:
    """Validates if required fields are present"""
    if "build-wasm" not in d_in.keys():
        typer.echo(
            f"ERROR: Missing table '[build-wasm]'\n"
            f"       {icpp_toml_path.resolve()}"
        )
        sys.exit(1)

    dwasm = d_in.get("build-wasm", {})
    if "canister" not in dwasm.keys():
        typer.echo(
            f"ERROR: Missing key 'canister' in '[build-wasm]' table.\n"
            f"       {icpp_toml_path.resolve()}"
        )
        sys.exit(1)

    if "did_path" not in dwasm.keys():
        typer.echo(
            f"ERROR: Missing key 'did_path' in '[build-wasm]' table.\n"
            f"       {icpp_toml_path.resolve()}"
        )
        sys.exit(1)


def read_build_wasm_table(d_in: Dict[Any, Any]) -> Dict[Any, Any]:
    """Reads and processes the '[build-wasm]' table."""
    d = {}
    d["canister"] = d_in.get("canister", "default")
    d["did_path"] = icpp_toml_path.parent / Path(d_in.get("did_path", "")).resolve()
    d["did_file"] = str(d["did_path"]) + " "

    icpp_toml_path_dir = icpp_toml_path.parent.resolve()
    d["cpp_include_dirs"] = [icpp_toml_path_dir, icpp_toml_path_dir / "src"]
    d["c_include_dirs"] = [icpp_toml_path_dir, icpp_toml_path_dir / "src"]
    read_build_table_common(d, d_in)

    return d


def read_build_native_table(d_in: Dict[Any, Any]) -> Dict[Any, Any]:
    """Reads and processes the '[build-native]' table."""
    d: Dict[Any, Any] = {}
    icpp_toml_path_dir = icpp_toml_path.parent.resolve()
    d["cpp_include_dirs"] = [icpp_toml_path_dir / "native"]
    d["c_include_dirs"] = [icpp_toml_path_dir, icpp_toml_path_dir / "native"]
    read_build_table_common(d, d_in)
    return d


def read_build_table_common(d: Dict[Any, Any], d_in: Dict[Any, Any]) -> None:
    """Reads the common fields of '[build-wasm]' & '[build-native]'"""

    # Helper function to expand wildcards
    def expand_paths(patterns: List[str]) -> List[Path]:
        paths: Set[Path] = set()  # Using a set to store unique paths
        for pattern in patterns:
            absolute_pattern = (icpp_toml_path.parent / pattern).resolve()
            paths.update(map(Path, glob.glob(str(absolute_pattern))))
        return list(paths)

    d["cpp_paths"] = expand_paths(d_in.get("cpp_paths", []))
    if "cpp_include_dirs" in d.keys():
        d["cpp_include_dirs"].extend(expand_paths(d_in.get("cpp_include_dirs", [])))
    else:
        d["cpp_include_dirs"] = d_in.get("cpp_include_dirs", [])
    d["cpp_header_paths"] = expand_paths(d_in.get("cpp_header_paths", []))
    d["cpp_compile_flags"] = d_in.get("cpp_compile_flags", [])
    d["cpp_link_flags"] = d_in.get("cpp_link_flags", [])

    d["c_paths"] = expand_paths(d_in.get("c_paths", []))
    if "c_include_dirs" in d.keys():
        d["c_include_dirs"].extend(expand_paths(d_in.get("c_include_dirs", [])))
    else:
        d["c_include_dirs"] = expand_paths(d_in.get("c_include_dirs", []))

    d["c_header_paths"] = expand_paths(d_in.get("c_header_paths", []))
    d["c_compile_flags"] = d_in.get("c_compile_flags", [])

    # if non-empty, these defaults will overwrite internal settings
    d["cpp_compile_flags_defaults"] = d_in.get("cpp_compile_flags_defaults", [])
    d["cpp_link_flags_defaults"] = d_in.get("cpp_link_flags_defaults", [])
    d["c_compile_flags_defaults"] = d_in.get("c_compile_flags_defaults", [])

    #
    # concurrent compiler uses a list of strings for the files to compile
    #
    d["cpp_files_list"] = [str(x) for x in d["cpp_paths"]]

    d["c_files_list"] = [str(x) for x in d["c_paths"]]

    #
    # all in one compiler uses a long string of strings, not a list
    #
    d["cpp_files"] = " ".join([str(x) for x in d["cpp_paths"]]) + " "
    d["cpp_include_flags"] = (
        " ".join(["-I " + str(path) for path in d["cpp_include_dirs"]]) + " "
    )
    d["cpp_header_files"] = " ".join([str(x) for x in d["cpp_header_paths"]]) + " "
    d["cpp_compile_flags_s"] = " ".join([str(x) for x in d["cpp_compile_flags"]]) + " "

    d["c_files"] = " ".join([str(x) for x in d["c_paths"]]) + " "
    d["c_include_flags"] = (
        " ".join(["-I " + str(path) for path in d["c_include_dirs"]]) + " "
    )
    d["c_header_files"] = " ".join([str(x) for x in d["c_header_paths"]]) + " "
    d["c_compile_flags_s"] = " ".join([str(x) for x in d["c_compile_flags"]]) + " "

    #
    # linker uses a long string of strings, not a list
    #
    d["cpp_link_flags_s"] = " ".join([str(x) for x in d["cpp_link_flags"]]) + " "

    #
    # overwriting defaults also uses strings
    #
    d["cpp_compile_flags_defaults_s"] = (
        " ".join([str(x) for x in d["cpp_compile_flags_defaults"]]) + " "
    )
    d["c_compile_flags_defaults_s"] = (
        " ".join([str(x) for x in d["c_compile_flags_defaults"]]) + " "
    )
    d["cpp_link_flags_defaults_s"] = (
        " ".join([str(x) for x in d["cpp_link_flags_defaults"]]) + " "
    )

    d["overwrite_default_CFLAGS"] = False
    d["overwrite_default_CPPFLAGS"] = False
    d["overwrite_default_LDFLAGS"] = False
    if len(d["c_compile_flags_defaults"]) > 0:
        d["overwrite_default_CFLAGS"] = True
    if len(d["cpp_compile_flags_defaults"]) > 0:
        d["overwrite_default_CPPFLAGS"] = True
    if len(d["cpp_link_flags_defaults"]) > 0:
        d["overwrite_default_LDFLAGS"] = True


#
# Read the top level icpp.toml
#

with open(icpp_toml_path, "rb") as f:
    data = tomllib.load(f)

validate(data)
build_wasm: Dict[Any, Any] = read_build_wasm_table(data.get("build-wasm", {}))
build_native: Dict[Any, Any] = read_build_native_table(data.get("build-native", {}))



"""__main__.py
The double underscores indicate that this file has a special meaning in Python.
When running a package (!) as a script with -m, 'python -m icpp', Python
executes the contents of the __main__.py file.

In other words, __main__.py acts as the entry point of our program and takes care of
the main flow, calling other parts as needed

reference: https://realpython.com/pypi-publish-python-package/
"""
import typer


# Instantiate the Typer app globally with app = typer.Typer().
# That way, we can decorate any function we want to call from the command line,
# using the @app.command() decorator.
app = typer.Typer()

# do some stuff with app
# pylint: disable = wrong-import-position unused-import
from icpp import options_main
from icpp import commands_get
from icpp import commands_init
from icpp import commands_build_wasm
from icpp import commands_build_native
from icpp import commands_install_wasi_sdk


def main() -> None:
    """Entry point of program"""
    app(prog_name="icpp")


#
# Always start it up or debug as a module:
#  python -m icpp.__main__
#
main()



"""
Defines the version for the compatible wasi-sdk compiler.

(-) Release scripts might modify this automatically...
(-) do not add anything but the version number of the wasi-sdk compiler here!

"""
__version__ = "wasi-sdk-20.0"



"""Handles 'icpp get-...' """

import typer
from icpp.__main__ import app
from icpp import config_default


@app.command()
def get_icpp_path() -> None:
    """Prints path to icpp folder.

    For use in your build commands.
    """
    typer.echo(config_default.ICPP_DIR)


@app.command()
def get_api_c_and_cpp_files() -> None:
    """Lists all C & C++ files of the IC API.

    For use in your build commands.
    """
    typer.echo(config_default.IC_C_FILES + config_default.IC_CPP_FILES)


@app.command()
def get_api_header_files() -> None:
    """Lists all header files of the IC API.

    For use in your build commands.
    """
    typer.echo(config_default.IC_HEADER_FILES)


@app.command()
def get_mock_ic_cpp_files() -> None:
    """Lists all C++ files of the native debug Mock IC.

    For use in your native debug build commands.
    """
    typer.echo(config_default.MOCKIC_CPP_FILES)


@app.command()
def get_mock_ic_header_files() -> None:
    """Lists all header files of the native debug Mock IC.

    For use in your native debug build commands.
    """
    typer.echo(config_default.MOCKIC_HEADER_FILES)



"""Handles 'icpp build-native' """
# pylint: disable=too-many-statements
import sys
import subprocess
import shutil
import concurrent.futures
import typer
from typing_extensions import Annotated

from icpp.__main__ import app
from icpp import config_default
from icpp.run_shell_cmd import run_shell_cmd
from icpp.run_dfx_cmd import run_dfx_cmd

from icpp.decorators import requires_native_compiler, requires_pro
from icpp.options_build import (
    to_compile_callback,
    option_to_compile_values_string,
    generate_bindings_callback,
    option_generate_bindings_values_string,
)

# options are: "none", "multi-threading"
CONCURRENCY = "multi-threading"


@app.command()
@requires_native_compiler()
@requires_pro("build-native")
def build_native(
    to_compile: Annotated[
        str,
        typer.Option(
            help=f"Files to compile {option_to_compile_values_string}.",
            callback=to_compile_callback,
        ),
    ] = "all",
    generate_bindings: Annotated[
        str,
        typer.Option(
            help=(
                f"Generate Javascript bindings "
                f"{option_generate_bindings_values_string}."
            ),
            callback=generate_bindings_callback,
        ),
    ] = "yes",
) -> None:
    """Builds a native debug executable with your systems' Clang compiler.

    Reads 'icpp.toml' file in the current folder, and uses: \n
    (-) C++ & C files from [build-native] AND [build-wasm]\n
    (-) Only the compile flags of [build-native] \n
    (-) Only the link flags of [build-native]
    """
    from icpp import icpp_toml  # pylint: disable = import-outside-toplevel

    build_path = icpp_toml.icpp_toml_path.parent / "build-native"
    typer.echo(f"Build folder: {build_path.resolve()}")

    if to_compile == "all":
        if build_path.exists():
            typer.echo("Deleting the build-native folder.")
            try:
                shutil.rmtree(build_path)
            except OSError as e:
                typer.echo(f"ERROR: During delete of {build_path}:\n {e.strerror}")

    if not build_path.exists():
        build_path.mkdir()

    cpp_files = icpp_toml.build_native["cpp_files"] + icpp_toml.build_wasm["cpp_files"]
    cpp_files_list = (
        icpp_toml.build_native["cpp_files_list"]
        + icpp_toml.build_wasm["cpp_files_list"]
    )
    cpp_include_flags = (
        icpp_toml.build_native["cpp_include_flags"]
        + icpp_toml.build_wasm["cpp_include_flags"]
    )
    cpp_compile_flags_s = icpp_toml.build_native["cpp_compile_flags_s"]

    c_files = icpp_toml.build_native["c_files"] + icpp_toml.build_wasm["c_files"]
    c_files_list = (
        icpp_toml.build_native["c_files_list"] + icpp_toml.build_wasm["c_files_list"]
    )
    c_include_flags = (
        icpp_toml.build_native["c_include_flags"]
        + icpp_toml.build_wasm["c_include_flags"]
    )
    c_compile_flags_s = icpp_toml.build_native["c_compile_flags_s"]

    cpp_compile_flags_defaults_s = config_default.NATIVE_CPPFLAGS
    cpp_link_flags_defaults_s = config_default.NATIVE_LDFLAGS
    c_compile_flags_defaults_s = config_default.NATIVE_CFLAGS

    if icpp_toml.build_native["overwrite_default_CPPFLAGS"]:
        cpp_compile_flags_defaults_s = icpp_toml.build_native[
            "cpp_compile_flags_defaults_s"
        ]
    if icpp_toml.build_native["overwrite_default_LDFLAGS"]:
        cpp_link_flags_defaults_s = icpp_toml.build_native["cpp_link_flags_defaults_s"]
    if icpp_toml.build_native["overwrite_default_CFLAGS"]:
        c_compile_flags_defaults_s = icpp_toml.build_native[
            "c_compile_flags_defaults_s"
        ]

    def cpp_compile_cmd_default() -> str:
        cmd = (
            f"{config_default.NATIVE_CPP} {config_default.NATIVE_CPP_REQUIRED_FLAGS} "
            f"{cpp_compile_flags_defaults_s} "
        )
        return cmd

    def c_compile_cmd_default() -> str:
        cmd = (
            f"{config_default.NATIVE_C} {config_default.NATIVE_C_REQUIRED_FLAGS} "
            f"{c_compile_flags_defaults_s} "
        )
        return cmd

    def cpp_compile_cmd_mine() -> str:
        cmd = (
            f"{cpp_compile_cmd_default()} "
            f"{cpp_include_flags} "
            f"{cpp_compile_flags_s} "
        )
        return cmd

    def c_compile_cmd_mine() -> str:
        cmd = f"{c_compile_cmd_default()} {c_include_flags} {c_compile_flags_s} "
        return cmd

    def cpp_compile_file_mine(file: str) -> None:
        cmd = f"{cpp_compile_cmd_mine()} -c {file}"
        typer.echo(cmd)
        run_shell_cmd(cmd, cwd=build_path)

    def c_compile_file_mine(file: str) -> None:
        cmd = f"{c_compile_cmd_mine()} -c {file}"
        typer.echo(cmd)
        run_shell_cmd(cmd, cwd=build_path)

    def cpp_compile_file_icpp(file: str) -> None:
        cmd = f"{cpp_compile_cmd_default()} -c {file}"
        typer.echo(cmd)
        run_shell_cmd(cmd, cwd=build_path)

    def c_compile_file_icpp(file: str) -> None:
        cmd = f"{c_compile_cmd_default()} -c {file}"
        typer.echo(cmd)
        run_shell_cmd(cmd, cwd=build_path)

    try:
        # ----------------------------------------------------------------------
        # compile 'mine' C++ files, if we have any
        if len(cpp_files.strip()) > 0:
            if CONCURRENCY == "multi-threading":
                typer.echo("--")
                typer.echo("Compiling your C++ files using multi-threading:")
                with concurrent.futures.ThreadPoolExecutor() as executor:
                    executor.map(cpp_compile_file_mine, cpp_files_list)
            else:
                cmd = f"{cpp_compile_cmd_mine()} -c {cpp_files}"

                typer.echo("--")
                typer.echo("Compiling your C++ files with command:")
                typer.echo(cmd)
                run_shell_cmd(cmd, cwd=build_path)

        # ----------------------------------------------------------------------
        # compile 'mine' C files, if we have any
        if len(c_files.strip()) > 0:
            if CONCURRENCY == "multi-threading":
                typer.echo("--")
                typer.echo("Compiling your C files using multi-threading:")
                with concurrent.futures.ThreadPoolExecutor() as executor:
                    executor.map(c_compile_file_mine, c_files_list)
            else:
                cmd = f"{c_compile_cmd_mine()} -c {c_files}"

                typer.echo("--")
                typer.echo("Compiling your C files with command:")
                typer.echo(cmd)
                run_shell_cmd(cmd, cwd=build_path)

        if to_compile == "all":
            # ----------------------------------------------------------------------
            # compile the C++ files of the Mock IC, if we have any
            if len(config_default.MOCKIC_CPP_FILES.strip()) > 0:
                if CONCURRENCY == "multi-threading":
                    typer.echo("--")
                    typer.echo(
                        "Compiling C++ files of the Mock IC using multi-threading:"
                    )
                    with concurrent.futures.ThreadPoolExecutor() as executor:
                        executor.map(
                            cpp_compile_file_icpp,
                            config_default.MOCKIC_CPP_FILES_LIST,
                        )
                else:
                    cmd = (
                        f"{cpp_compile_cmd_default()} "
                        f"-c {config_default.MOCKIC_CPP_FILES}"
                    )

                    typer.echo("--")
                    typer.echo("Compiling C++ files of the Mock IC with command:")
                    typer.echo(cmd)
                    run_shell_cmd(cmd, cwd=build_path)

            # ----------------------------------------------------------------------
            # compile the C files of the Mock IC, if we have any
            if len(config_default.MOCKIC_C_FILES.strip()) > 0:
                if CONCURRENCY == "multi-threading":
                    typer.echo("--")
                    typer.echo(
                        "Compiling C++ files of the Mock IC using multi-threading:"
                    )
                    with concurrent.futures.ThreadPoolExecutor() as executor:
                        executor.map(
                            c_compile_file_icpp,
                            config_default.MOCKIC_C_FILES_LIST,
                        )
                else:
                    cmd = (
                        f"{c_compile_cmd_default()} -c {config_default.MOCKIC_C_FILES}"
                    )

                    typer.echo("--")
                    typer.echo("Compiling C files of the Mock IC with command:")
                    typer.echo(cmd)
                    run_shell_cmd(cmd, cwd=build_path)

        # ----------------------------------------------------------------------
        # link it into a native executable
        executable = "mockic.exe"
        cmd = (
            f"{config_default.NATIVE_CPP} "
            f"{config_default.NATIVE_CPP_REQUIRED_FLAGS} "
            f"{cpp_link_flags_defaults_s} "
            f"*.o -o {executable}"
        )

        typer.echo("--")
        typer.echo("Linking it into a native executable with command:")
        typer.echo(cmd)
        run_shell_cmd(cmd, cwd=build_path)

    except subprocess.CalledProcessError as e:
        sys.exit(e.returncode)
    except Exception as e:  # pylint: disable=broad-except
        typer.echo(
            f"ERROR: something went wrong in module 'commands_build_native.py' \n "
            f"Exception is: \n{e}"
        )
        sys.exit(1)

    # ----------------------------------------------------------------------
    # All done
    typer.echo("--")
    typer.echo("All done building the native debug executable:")
    typer.echo(f"{build_path.resolve()}/{executable}")
    try:
        typer.echo("(✔)You can run it from the command line")
        typer.echo("(✔)You can debug it with VS Code + CodeLLDB")
    except UnicodeEncodeError:
        typer.echo("(-)You can run it from the command line")
        typer.echo("(-)You can debug it with VS Code + CodeLLDB")

    # ----------------------------------------------------------------------
    typer.echo("--")
    if generate_bindings.lower() == "no":
        typer.echo("Skipping generation of Javascript bindings from your .did file.")
    else:
        typer.echo("Generating Javascript bindings from your .did file:")
        declarations_path = icpp_toml.icpp_toml_path.parent / "src/declarations"
        typer.echo(f"{declarations_path.resolve()}/{icpp_toml.build_wasm['canister']}")
        run_dfx_cmd("generate")
        try:
            typer.echo("✔️")
        except UnicodeEncodeError:
            typer.echo(" ")

    # ----------------------------------------------------------------------
    typer.echo("--")



"""Functions to be used in a pytest test"""

import subprocess
import json
import platform
from pathlib import Path
from typing import Optional, Any, Dict
import pytest  # pylint: disable=unused-import

from icpp.run_shell_cmd import run_shell_cmd
from icpp import pro

DFX = "dfx"
RUN_IN_POWERSHELL = False
if platform.win32_ver()[0]:
    DFX = "wsl --% dfx"
    RUN_IN_POWERSHELL = True


def call_canister_api(
    *,
    dfx_json_path: Path,
    canister_name: str,
    canister_method: str,
    canister_argument: Optional[str] = None,
    canister_input: str = "idl",
    canister_output: str = "idl",
    network: str = "local",
    quiet: str = "-qq",  # limits dfx to errors only
    timeout_seconds: Optional[int] = None,
) -> str:
    """Calls a canister method"""
    pro.exit_if_not_pro("smoketesting with pytest")

    # verify that canister_name is part of dfx.json
    with open(dfx_json_path, "rb") as f:
        data = json.load(f)

    if canister_name not in data["canisters"].keys():
        pytest.fail(
            f"ERROR: canister '{canister_name}' not defined in '{str(dfx_json_path)}'"
        )

    arg = (
        f"{DFX} "
        f" canister "
        f" --network {network} "
        f" {quiet} "
        f" call "
        f" --type {canister_input} "
        f" --output {canister_output} "
        f" {canister_name} "
        f" {canister_method} "
    )

    if canister_argument is None:
        arg += " '()' "
    else:
        arg += f" '{canister_argument}' "

    try:
        # response = run_shell_cmd(arg, cwd=Path(dfx_json_path).parent)
        response = run_shell_cmd(
            arg,
            capture_output=True,
            cwd=Path(dfx_json_path).parent,
            timeout_seconds=timeout_seconds,
            run_in_powershell=RUN_IN_POWERSHELL,
        )
        response = response.rstrip("\n")
    except subprocess.CalledProcessError as e:
        if "Error: Failed to determine id for canister" in e.output:
            msg = (
                "\n\n"
                f"FAIL: Call to api '{canister_method}' of canister "
                f"'{canister_name}'\n\n"
                "*******************************************\n"
                "*** Failed to determine id for canister ***\n"
                "*** Deploy the canister first with:     ***\n"
                f"***  {DFX} deploy                       ***\n"
                "*******************************************"
            )
            pytest.exit(msg)
        else:
            response = (
                f"Failed call to api '{canister_method}' of canister '{canister_name}':"
                f"{e.output}"
            )

    return response


def dict_to_candid_text(d: Dict[Any, Any]) -> str:
    """Serializes dict to Candid text to send it over the wire"""
    pro.exit_if_not_pro("smoketesting with pytest")
    # pull it through twice, so all " are escaped into \"
    return json.dumps(json.dumps(d))


def network_status(network: str) -> str:
    """Returns the network status."""
    pro.exit_if_not_pro("smoketesting with pytest")
    arg = f"{DFX} ping {network}"
    try:
        response = run_shell_cmd(
            arg, capture_output=False, run_in_powershell=RUN_IN_POWERSHELL
        )
    except subprocess.CalledProcessError:
        if network == "local":
            msg = (
                "\n"
                "***************************************\n"
                "*** The local network is not up     ***\n"
                "*** Please start it first with:     ***\n"
                f"***  {DFX} start --clean --background ***\n"
                "***************************************"
            )
        else:
            msg = (
                "\n"
                "*******************************\n"
                "*** The IC is not up        ***\n"
                "*** Please try again later  ***\n"
                "*******************************"
            )
        pytest.exit(msg)

    return response


def get_identity() -> str:
    """Returns the current dfx identity."""
    pro.exit_if_not_pro("smoketesting with pytest")
    arg = f"{DFX} identity whoami "
    try:
        identity = run_shell_cmd(
            arg,
            capture_output=True,
            timeout_seconds=5,
            run_in_powershell=RUN_IN_POWERSHELL,
        )
        identity = identity.rstrip("\n")
    except subprocess.CalledProcessError as e:
        pytest.fail(f"ERROR: command {arg} failed with error:\n{e.output}")

    return identity


def set_identity(identity: str) -> None:
    """Sets the dfx identity."""
    pro.exit_if_not_pro("smoketesting with pytest")
    arg = f"{DFX} identity use {identity}"
    try:
        run_shell_cmd(arg, run_in_powershell=RUN_IN_POWERSHELL)
    except subprocess.CalledProcessError as e:
        pytest.fail(f"ERROR: command {arg} failed with error:\n{e.output}")


def get_principal() -> str:
    """Returns the principal of the current dfx identity."""
    pro.exit_if_not_pro("smoketesting with pytest")
    arg = f"{DFX} identity get-principal "
    try:
        principal = run_shell_cmd(
            arg,
            capture_output=True,
            timeout_seconds=5,
            run_in_powershell=RUN_IN_POWERSHELL,
        )
        principal = principal.rstrip("\n")
    except subprocess.CalledProcessError as e:
        pytest.fail(f"ERROR: command {arg} failed with error:\n{e.output}")

    return principal



"""Run a dfx command"""
import sys
import subprocess
import platform
from typing import Optional
from icpp.run_shell_cmd import run_shell_cmd


DFX = "dfx"
RUN_IN_POWERSHELL = False
if platform.win32_ver()[0]:
    DFX = "wsl --% dfx"
    RUN_IN_POWERSHELL = True


def run_dfx_cmd(args: str) -> Optional[str]:
    """Runs dfx command as a subprocess"""
    try:
        cmd = f"{DFX} {args} "
        return run_shell_cmd(
            cmd,
            capture_output=True,
            run_in_powershell=RUN_IN_POWERSHELL,
        ).rstrip("\n")
    except subprocess.CalledProcessError as e:
        print(f"Failed dfx command: '{cmd}' with error: \n{e.output}")
        sys.exit(1)
    return None



"""Callback functions of our icpp typer app"""
from typing import Optional
import typer
from icpp.__main__ import app
from icpp import __version__, __version_wasi_sdk__


def version_callback(value: bool) -> None:
    """Prints version of icpp & wasi-sdk.

    When the user issues the command `icpp --version`, this callback is called.
    """
    if value:
        typer.echo(f"icpp-pro version: {__version__}")
        typer.echo(f"wasi-sdk version: {__version_wasi_sdk__}")
        raise typer.Exit()


@app.callback()
# pylint: disable=unused-argument
def main_callback(
    # https://typer.tiangolo.com/tutorial/options/version/
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        help="Prints version of icpp & wasi-sdk.",
        callback=version_callback,
        is_eager=True,
    ),
) -> None:
    """A development platform for C++ smart contracts of the Internet Computer"""



"""Handles validation of --compile option """
from typing import Optional
import typer
from icpp import pro

option_to_compile_values = ["all", "mine"]
option_to_compile_values_string = f"[{'/'.join(option_to_compile_values)}]"

option_generate_bindings_values = ["yes", "no"]
option_generate_bindings_values_string = (
    f"[{'/'.join(option_generate_bindings_values)}]"
)


def to_compile_callback(ctx: typer.Context, value: str) -> Optional[str]:
    """--to-compile [...]"""
    # Handle auto complete
    if ctx.resilient_parsing:
        return None

    if value not in option_to_compile_values:
        msg = f"'{value}'\nValid values are: {option_to_compile_values_string}"
        raise typer.BadParameter(msg)

    if value != "all":
        pro.exit_if_not_pro(f"--to-compile {value}")

    return value


def generate_bindings_callback(ctx: typer.Context, value: str) -> Optional[str]:
    """--generate-bindings [...]"""
    # Handle auto complete
    if ctx.resilient_parsing:
        return None

    if value not in option_generate_bindings_values:
        msg = f"'{value}'\nValid values are: {option_generate_bindings_values_string}"
        raise typer.BadParameter(msg)

    return value



"""
Set the default config variables
"""
import sys
import platform
from pathlib import Path
import getpass
import icpp_candid  # pylint: disable = unused-import
from icpp import __version_wasi_sdk__

# The OS we're running on - https://stackoverflow.com/a/1857/5480536

OS_SYSTEM = platform.system()
OS_PROCESSOR = platform.processor()


#######################################################################
# The compatible wasi-sdk compiler


def get_wasi_sdk_os_name() -> str:
    """Returns the os name used for the releases build by the CI/CD of the wasi-sdk
    repo, for the current OS."""

    if OS_SYSTEM == "Linux":
        return "-linux"

    if OS_SYSTEM == "Darwin":
        return "-macos"

    if OS_SYSTEM == "Windows":
        # Naming changed with wasi-sdk 20
        return ".m-mingw"

    return "unknown"


def get_wasi_sdk_untar_dir_name() -> str:
    """Returns the dir name after untarring, for the current OS."""

    if OS_SYSTEM == "Linux":
        return __version_wasi_sdk__

    if OS_SYSTEM == "Darwin":
        return __version_wasi_sdk__

    if OS_SYSTEM == "Windows":
        # Naming changed with wasi-sdk 20
        return __version_wasi_sdk__ + "+m"

    return "unknown"


WASI_SDK_OS_NAME = get_wasi_sdk_os_name()
WASI_SDK_URL = (
    f"https://github.com/WebAssembly/wasi-sdk/releases/download/"
    f"{__version_wasi_sdk__.split('.',1)[0]}/"
    f"{__version_wasi_sdk__}{WASI_SDK_OS_NAME}.tar.gz"
)

ICPP_ROOT = Path.home() / ".icpp"
ICPP_ROOT_UNTAR_DIR = ICPP_ROOT / f"{get_wasi_sdk_untar_dir_name()}"
ICPP_ROOT_COMPILER = ICPP_ROOT / f"{__version_wasi_sdk__}"

USER = getpass.getuser()

ICPP_PATH = Path(sys.modules["icpp"].__path__[0])
ICPP_DIR = str(ICPP_PATH)

ICPP_CANDID_PATH = Path(sys.modules["icpp_candid"].__path__[0])
ICPP_CANDID_DIR = str(ICPP_CANDID_PATH)

########################################################################
# The greet canister created by `icpp init`
#
ICPP_INIT_PATH = ICPP_PATH / "canisters/greet"
LOCAL_INIT_PATH = Path("./greet")

########################################################################
# build-wasm
#
IC_C_FILES_LIST = [
    str(x)
    for x in list(ICPP_PATH.glob("ic/ic0/*.c"))
    + list(ICPP_PATH.glob("ic/wasi_sdk_traps/*.c"))
    + list(ICPP_PATH.glob("ic/canister/*.c"))
    + list(ICPP_PATH.glob("ic/icapi/*.c"))
    + list(ICPP_CANDID_PATH.glob("candid/*.c"))
]
IC_CPP_FILES_LIST = [
    str(x)
    for x in list(ICPP_PATH.glob("ic/ic0/*.cpp"))
    + list(ICPP_PATH.glob("ic/canister/*.cpp"))
    + list(ICPP_PATH.glob("ic/icapi/*.cpp"))
    + list(ICPP_PATH.glob("ic/pro/*.cpp"))
    + list(ICPP_CANDID_PATH.glob("candid/*.cpp"))
    + list(ICPP_CANDID_PATH.glob("hooks/*.cpp"))
    + list(ICPP_CANDID_PATH.glob("vendors/hash-library/*.cpp"))
]
IC_HEADER_FILES_LIST = [
    str(x)
    for x in list(ICPP_PATH.glob("ic/ic0/*.h"))
    + list(ICPP_PATH.glob("ic/wasi_sdk_traps/*.h"))
    + list(ICPP_PATH.glob("ic/canister/*.h"))
    + list(ICPP_PATH.glob("ic/icapi/*.h"))
    + list(ICPP_PATH.glob("ic/pro/*.h"))
    + list(ICPP_CANDID_PATH.glob("candid/*.h"))
    + list(ICPP_CANDID_PATH.glob("hooks/*.h"))
    + list(ICPP_CANDID_PATH.glob("vendors/hash-library/*.h"))
    + list(ICPP_CANDID_PATH.glob("vendors/cppcodec/*.hpp"))
    + list(ICPP_CANDID_PATH.glob("vendors/cppcodec/*/*.hpp"))
]

IC_C_FILES = " ".join(IC_C_FILES_LIST) + " "
IC_CPP_FILES = " ".join(IC_CPP_FILES_LIST) + " "
IC_HEADER_FILES = " ".join(IC_HEADER_FILES_LIST) + " "

SYSROOT = ICPP_ROOT_COMPILER / "share/wasi-sysroot"

WASM_C_REQUIRED_FLAGS = (
    f" --target=wasm32-wasi --sysroot {SYSROOT} "
    f" -I {ICPP_DIR}/ic/canister -I {ICPP_DIR}/ic/icapi "
    f" -I {ICPP_DIR}/ic/pro -I {ICPP_DIR}/ic/ic0 "
    f" -I {ICPP_CANDID_PATH}/candid -I {ICPP_CANDID_PATH}/hooks "
    f" -I {ICPP_CANDID_PATH}/vendors "
)
WASM_CPP_REQUIRED_FLAGS = WASM_C_REQUIRED_FLAGS + " -std=c++20 "

WASM_C = ICPP_ROOT_COMPILER / "bin/clang"
WASM_CPP = ICPP_ROOT_COMPILER / "bin/clang++"
WASM_CFLAGS = (
    " -O3 -flto -fno-exceptions -fvisibility=hidden -D NDEBUG -D ICPP_VERBOSE=0 "
)
WASM_CPPFLAGS = WASM_CFLAGS
WASM_LDFLAGS = (
    " -nostartfiles -Wl,--no-entry -Wl,--lto-O3 -Wl,--strip-all "
    " -Wl,--strip-debug -Wl,--stack-first -Wl,--export-dynamic "
)

# Utilities we can use too
CLANG_FORMAT = ICPP_ROOT_COMPILER / "bin/clang-format"
CLANG_TIDY = ICPP_ROOT_COMPILER / "bin/clang-tidy"

########################################################################
# build-native
#
MOCKIC_C_FILES_LIST = [
    str(x)
    for x in list(ICPP_PATH.glob("ic/ic0mock/*.c"))
    + list(ICPP_PATH.glob("ic/canister/*.c"))
    + list(ICPP_PATH.glob("ic/icapi/*.c"))
    + list(ICPP_CANDID_PATH.glob("candid/*.c"))
]

MOCKIC_CPP_FILES_LIST = [
    str(x)
    for x in list(ICPP_PATH.glob("ic/ic0mock/*.cpp"))
    + list(ICPP_PATH.glob("ic/canister/*.cpp"))
    + list(ICPP_PATH.glob("ic/icapi/*.cpp"))
    + list(ICPP_PATH.glob("ic/pro/*.cpp"))
    + list(ICPP_CANDID_PATH.glob("candid/*.cpp"))
    + list(ICPP_CANDID_PATH.glob("hooks/*.cpp"))
    + list(ICPP_CANDID_PATH.glob("vendors/hash-library/*.cpp"))
]
MOCKIC_HEADER_FILES_LIST = [
    str(x)
    for x in list(ICPP_PATH.glob("ic/ic0mock/*.h"))
    + list(ICPP_PATH.glob("ic/canister/*.h"))
    + list(ICPP_PATH.glob("ic/icapi/*.h"))
    + list(ICPP_PATH.glob("ic/pro/*.h"))
    + list(ICPP_CANDID_PATH.glob("candid/*.h"))
    + list(ICPP_CANDID_PATH.glob("hooks/*.h"))
    + list(ICPP_CANDID_PATH.glob("vendors/hash-library/*.h"))
    + list(ICPP_CANDID_PATH.glob("vendors/cppcodec/*.hpp"))
    + list(ICPP_CANDID_PATH.glob("vendors/cppcodec/*/*.hpp"))
]

MOCKIC_C_FILES = " ".join(MOCKIC_C_FILES_LIST) + " "
MOCKIC_CPP_FILES = " ".join(MOCKIC_CPP_FILES_LIST) + " "
MOCKIC_HEADER_FILES = " ".join(MOCKIC_HEADER_FILES_LIST) + " "

NATIVE_COMPILER = "Clang"
NATIVE_C = "clang"
NATIVE_CPP = "clang++"

NATIVE_C_REQUIRED_FLAGS = (
    f" -I {ICPP_DIR}/ic/canister -I {ICPP_DIR}/ic/icapi "
    f" -I {ICPP_DIR}/ic/pro -I {ICPP_DIR}/ic/ic0mock "
    f" -I {ICPP_CANDID_PATH}/candid -I {ICPP_CANDID_PATH}/hooks "
    f" -I {ICPP_CANDID_PATH}/vendors "
)
NATIVE_CPP_REQUIRED_FLAGS = NATIVE_C_REQUIRED_FLAGS + " -std=c++20 "

NATIVE_CFLAGS = " -g -D ICPP_VERBOSE=0 "
NATIVE_CPPFLAGS = NATIVE_CFLAGS
NATIVE_LDFLAGS = " "



"""pytest fixtures provided by icpp
   https://docs.pytest.org/en/latest/fixture.html
"""

from typing import Any, Generator, Dict
import pytest


from icpp.smoketest import network_status, get_identity, set_identity, get_principal
from icpp import pro


def pytest_addoption(parser: Any) -> None:
    """Adds options: `pytest --network=[local/ic] `"""
    pro.exit_if_not_pro("smoketesting with pytest")
    parser.addoption(
        "--network",
        action="store",
        default="local",
        help="The network to use: local or ic",
    )


###########################################################
# Some fixtures that always run, prior to the actual tests
# They verify everything is OK with the network & identity


@pytest.fixture(scope="session", autouse=True)
def network(request: Any) -> Any:
    """A fixture that verifies the network is up & returns the name."""
    pro.exit_if_not_pro("smoketesting with pytest")
    network_ = request.config.getoption("--network")
    network_status(network_)
    return network_


@pytest.fixture(scope="session", autouse=True)
def identity() -> Any:
    """A fixture that returns the name of the used identity."""
    pro.exit_if_not_pro("smoketesting with pytest")
    identity_ = get_identity()
    if identity_.startswith("ERROR"):
        raise Exception(identity_)
    return identity_


@pytest.fixture(scope="session", autouse=True)
def principal() -> Any:
    """A fixture that returns the principal of the used identity."""
    pro.exit_if_not_pro("smoketesting with pytest")
    principal_ = get_principal()
    if principal_.startswith("ERROR"):
        if "Please enter the passphrase for your identity" in principal_:
            msg = (
                f"Identity '{get_identity()}' uses a passphrase. "
                f"Use identity created with '--storage-mode plaintext'!"
            )
            raise Exception(msg)
        raise Exception(principal_)
    return principal_


####################################################################
# Fixtures to run a function with the anonymous or default identity


def handle_identity(identity_to_set: str) -> Generator[Dict[str, str], None, None]:
    """A fixture that sets the dfx identity."""
    pro.exit_if_not_pro("smoketesting with pytest")
    identity_before_test = get_identity()
    set_identity(identity_to_set)
    user = {"identity": get_identity(), "principal": get_principal()}
    yield user
    set_identity(identity_before_test)


@pytest.fixture(scope="function")
def identity_anonymous() -> Generator[Dict[str, str], None, None]:
    """A fixture that sets the dfx identity to anonymous."""
    pro.exit_if_not_pro("smoketesting with pytest")
    yield from handle_identity("anonymous")


@pytest.fixture(scope="function")
def identity_default() -> Generator[Dict[str, str], None, None]:
    """A fixture that sets the dfx identity to default."""
    pro.exit_if_not_pro("smoketesting with pytest")
    yield from handle_identity("default")



"""Provides decorators for the icpp commands

When creating a new decorator, add it to `icpp/__init__.py`,
so they become available in the package namespace, and you
can then use them as follows:

```
import icpp

@icpp.requires_wasi_sdk()
@app.command()
def build:
    ...
```

See: https://realpython.com/primer-on-python-decorators/
     https://mypy.readthedocs.io/en/stable/generics.html#declaring-decorators
"""
import sys
import shutil
from typing import Any, Callable, TypeVar, Optional
from functools import wraps
import typer
from icpp import config_default, pro
from icpp.commands_install_wasi_sdk import is_wasi_sdk_installed, install_wasi_sdk


F = TypeVar("F", bound=Callable[..., Any])


def requires_pro(capability: Optional[str] = None) -> Callable[[F], F]:
    """Decorates a command that requires icpp-pro.

    Exit if not running a licensed icpp-pro.
    """

    def decorator(f: F) -> Any:
        @wraps(f)
        def decorated(*args: Any, **kwargs: Any) -> Any:
            pro.exit_if_not_pro(capability)
            return f(*args, **kwargs)

        return decorated

    return decorator


def requires_wasi_sdk() -> Callable[[F], F]:
    """Decorates a command that requires the wasi-sdk compiler.

    If the wasi-sdk is not installed, it will do that first.
    """

    def decorator(f: F) -> Any:
        @wraps(f)
        def decorated(*args: Any, **kwargs: Any) -> Any:
            if not is_wasi_sdk_installed():
                typer.echo("--")
                typer.echo("The wasi-sdk is not installed. Will do that first.")
                install_wasi_sdk()
                typer.echo("Now that the wasi-sdk is installed, we can build things.")

            return f(*args, **kwargs)

        return decorated

    return decorator


def requires_native_compiler() -> Callable[[F], F]:
    """Decorates a command that requires a native compiler.

    If the native compiler is not installed, it will exit.
    """

    def decorator(f: F) -> Any:
        @wraps(f)
        def decorated(*args: Any, **kwargs: Any) -> Any:
            exit_if_native_compiler_not_installed()
            return f(*args, **kwargs)

        return decorated

    return decorator


def exit_if_native_compiler_not_installed() -> None:
    """Exit if `clang` or `clang++` not available"""
    no_c = shutil.which(config_default.NATIVE_C) is None
    no_cpp = shutil.which(config_default.NATIVE_CPP) is None
    if no_c or no_cpp:
        typer.echo("--")
        typer.echo("ERROR: The native compiler is not installed")
        if no_c:
            try:
                typer.echo(f"       (✗) could not find '{config_default.NATIVE_C}'")
            except UnicodeEncodeError:
                typer.echo(f"       (ERROR) could not find '{config_default.NATIVE_C}'")
        else:
            try:
                typer.echo(f"       (✔) found '{config_default.NATIVE_C}'")
            except UnicodeEncodeError:
                typer.echo(f"       (OK) found '{config_default.NATIVE_C}'")
        if no_cpp:
            try:
                typer.echo(f"       (✗) could not find '{config_default.NATIVE_CPP}'")
            except UnicodeEncodeError:
                typer.echo(
                    f"       (ERROR) could not find '{config_default.NATIVE_CPP}'"
                )
        else:
            try:
                typer.echo(f"       (✔) found '{config_default.NATIVE_CPP}'")
            except UnicodeEncodeError:
                typer.echo(f"       (OK) found '{config_default.NATIVE_CPP}'")
        sys.exit(1)



"""Handles 'icpp build-wasm' """
# pylint: disable = too-many-statements
import sys
import subprocess
import shutil
import concurrent.futures
import typer
from typing_extensions import Annotated

from icpp.__main__ import app

from icpp import config_default
from icpp.run_shell_cmd import run_shell_cmd
from icpp.run_dfx_cmd import run_dfx_cmd

from icpp.decorators import requires_wasi_sdk
from icpp.options_build import (
    to_compile_callback,
    option_to_compile_values_string,
    generate_bindings_callback,
    option_generate_bindings_values_string,
)

# options are: "none", "multi-threading"
CONCURRENCY = "multi-threading"


@app.command()
@requires_wasi_sdk()
def build_wasm(
    to_compile: Annotated[
        str,
        typer.Option(
            help=f"Files to compile {option_to_compile_values_string}.",
            callback=to_compile_callback,
        ),
    ] = "all",
    generate_bindings: Annotated[
        str,
        typer.Option(
            help=(
                f"Generate Javascript bindings "
                f"{option_generate_bindings_values_string}."
            ),
            callback=generate_bindings_callback,
        ),
    ] = "yes",
) -> None:
    """Builds the wasm for a canister, using the wasi-sdk compiler.

    Reads icpp.toml in the current folder; Compiles & builds a wasm file.
    """
    from icpp import icpp_toml  # pylint: disable = import-outside-toplevel

    build_path = icpp_toml.icpp_toml_path.parent / "build"
    typer.echo(f"Build folder: {build_path.resolve()}")

    if to_compile == "all":
        if build_path.exists():
            typer.echo("Deleting the build folder.")
            try:
                shutil.rmtree(build_path)
            except OSError as e:
                typer.echo(f"ERROR: During delete of {build_path}:\n {e.strerror}")

    if not build_path.exists():
        build_path.mkdir()

    cpp_files = icpp_toml.build_wasm["cpp_files"]
    cpp_files_list = icpp_toml.build_wasm["cpp_files_list"]
    cpp_include_flags = icpp_toml.build_wasm["cpp_include_flags"]
    cpp_compile_flags_s = icpp_toml.build_wasm["cpp_compile_flags_s"]

    c_files = icpp_toml.build_wasm["c_files"]
    c_files_list = icpp_toml.build_wasm["c_files_list"]
    c_include_flags = icpp_toml.build_wasm["c_include_flags"]
    c_compile_flags_s = icpp_toml.build_wasm["c_compile_flags_s"]

    cpp_compile_flags_defaults_s = config_default.WASM_CPPFLAGS
    cpp_link_flags_defaults_s = config_default.WASM_LDFLAGS
    c_compile_flags_defaults_s = config_default.WASM_CFLAGS

    if icpp_toml.build_wasm["overwrite_default_CPPFLAGS"]:
        cpp_compile_flags_defaults_s = icpp_toml.build_wasm[
            "cpp_compile_flags_defaults_s"
        ]
    if icpp_toml.build_wasm["overwrite_default_LDFLAGS"]:
        cpp_link_flags_defaults_s = icpp_toml.build_wasm["cpp_link_flags_defaults_s"]
    if icpp_toml.build_wasm["overwrite_default_CFLAGS"]:
        c_compile_flags_defaults_s = icpp_toml.build_wasm["c_compile_flags_defaults_s"]

    def cpp_compile_cmd_default() -> str:
        cmd = (
            f"{config_default.WASM_CPP} {config_default.WASM_CPP_REQUIRED_FLAGS} "
            f"{cpp_compile_flags_defaults_s} "
        )
        return cmd

    def c_compile_cmd_default() -> str:
        cmd = (
            f"{config_default.WASM_C} {config_default.WASM_C_REQUIRED_FLAGS} "
            f"{c_compile_flags_defaults_s} "
        )
        return cmd

    def cpp_compile_cmd_mine() -> str:
        cmd = (
            f"{cpp_compile_cmd_default()} "
            f"{cpp_include_flags} "
            f"{cpp_compile_flags_s} "
        )
        return cmd

    def c_compile_cmd_mine() -> str:
        cmd = f"{c_compile_cmd_default()} {c_include_flags} {c_compile_flags_s} "
        return cmd

    def cpp_compile_file_mine(file: str) -> None:
        cmd = f"{cpp_compile_cmd_mine()} -c {file}"
        typer.echo(cmd)
        run_shell_cmd(cmd, cwd=build_path)

    def c_compile_file_mine(file: str) -> None:
        cmd = f"{c_compile_cmd_mine()} -c {file}"
        typer.echo(cmd)
        run_shell_cmd(cmd, cwd=build_path)

    def cpp_compile_file_icpp(file: str) -> None:
        cmd = f"{cpp_compile_cmd_default()} -c {file}"
        typer.echo(cmd)
        run_shell_cmd(cmd, cwd=build_path)

    def c_compile_file_icpp(file: str) -> None:
        cmd = f"{c_compile_cmd_default()} -c {file}"
        typer.echo(cmd)
        run_shell_cmd(cmd, cwd=build_path)

    try:
        # ----------------------------------------------------------------------
        # compile 'mine' C++ files, if we have any
        if len(cpp_files.strip()) > 0:
            if CONCURRENCY == "multi-threading":
                typer.echo("--")
                typer.echo("Compiling your C++ files using multi-threading:")
                with concurrent.futures.ThreadPoolExecutor() as executor:
                    executor.map(cpp_compile_file_mine, cpp_files_list)
            else:
                cmd = f"{cpp_compile_cmd_mine()} -c {cpp_files}"

                typer.echo("--")
                typer.echo("Compiling your C++ files with command:")
                typer.echo(cmd)
                run_shell_cmd(cmd, cwd=build_path)

        # ----------------------------------------------------------------------
        # compile 'mine' C files, if we have any
        if len(c_files.strip()) > 0:
            if CONCURRENCY == "multi-threading":
                typer.echo("--")
                typer.echo("Compiling your C files using multi-threading:")
                with concurrent.futures.ThreadPoolExecutor() as executor:
                    executor.map(c_compile_file_mine, c_files_list)
            else:
                cmd = f"{c_compile_cmd_mine()} -c {c_files}"

                typer.echo("--")
                typer.echo("Compiling your C files with command:")
                typer.echo(cmd)
                run_shell_cmd(cmd, cwd=build_path)

        if to_compile == "all":
            # ----------------------------------------------------------------------
            # compile the C++ files of the IC API, if we have any
            if len(config_default.IC_CPP_FILES.strip()) > 0:
                if CONCURRENCY == "multi-threading":
                    typer.echo("--")
                    typer.echo(
                        "Compiling C++ files of the IC_API using multi-threading:"
                    )
                    with concurrent.futures.ThreadPoolExecutor() as executor:
                        executor.map(
                            cpp_compile_file_icpp,
                            config_default.IC_CPP_FILES_LIST,
                        )
                else:
                    cmd = (
                        f"{cpp_compile_cmd_default()} "
                        f"-c {config_default.IC_CPP_FILES}"
                    )

                    typer.echo("--")
                    typer.echo("Compiling C++ files of the IC_API with command:")
                    typer.echo(cmd)
                    run_shell_cmd(cmd, cwd=build_path)

            # ----------------------------------------------------------------------
            # compile the C files of the IC API, if we have any
            if len(config_default.IC_C_FILES.strip()) > 0:
                if CONCURRENCY == "multi-threading":
                    typer.echo("--")
                    typer.echo(
                        "Compiling C++ files of the IC_API using multi-threading:"
                    )
                    with concurrent.futures.ThreadPoolExecutor() as executor:
                        executor.map(
                            c_compile_file_icpp,
                            config_default.IC_C_FILES_LIST,
                        )
                else:
                    cmd = f"{c_compile_cmd_default()} -c {config_default.IC_C_FILES}"

                    typer.echo("--")
                    typer.echo("Compiling C files of the IC_API with command:")
                    typer.echo(cmd)
                    run_shell_cmd(cmd, cwd=build_path)

        # ----------------------------------------------------------------------
        # link it into a wasm
        cmd = (
            f"{config_default.WASM_CPP} "
            f"{config_default.WASM_CPP_REQUIRED_FLAGS} "
            f"{cpp_link_flags_defaults_s} "
            f"*.o -o {icpp_toml.build_wasm['canister']}.wasm"
        )

        typer.echo("--")
        typer.echo("Linking it into a wasm with command:")
        typer.echo(cmd)
        run_shell_cmd(cmd, cwd=build_path)

    except subprocess.CalledProcessError as e:
        sys.exit(e.returncode)
    except Exception as e:  # pylint: disable=broad-except
        typer.echo(
            f"ERROR: something went wrong in module 'commands_build_wasm.py' \n "
            f"Exception is: \n{e}"
        )
        sys.exit(1)

    # ----------------------------------------------------------------------
    # copy did into build folder
    did_path = icpp_toml.build_wasm["did_path"]
    typer.echo("--")
    if not did_path.exists():
        typer.echo(
            f"ERROR: Cannot find the `did_path` '{did_path}' \n"
            f"       Please update your icpp.toml"
        )
        sys.exit(1)

    typer.echo("Copying your Candid file to the build folder...")
    shutil.copy(did_path, build_path)

    # ----------------------------------------------------------------------
    typer.echo("--")
    typer.echo("All done building the canister Wasm module:")
    typer.echo(f"{build_path.resolve()}/{icpp_toml.build_wasm['canister']}.wasm")
    try:
        typer.echo("✔️")
    except UnicodeEncodeError:
        typer.echo(" ")

    # ----------------------------------------------------------------------
    typer.echo("--")
    if generate_bindings.lower() == "no":
        typer.echo("Skipping generation of Javascript bindings from your .did file.")
    else:
        typer.echo("Generating Javascript bindings from your .did file:")
        declarations_path = icpp_toml.icpp_toml_path.parent / "src/declarations"
        typer.echo(f"{declarations_path.resolve()}/{icpp_toml.build_wasm['canister']}")
        run_dfx_cmd("generate")
        try:
            typer.echo("✔️")
        except UnicodeEncodeError:
            typer.echo(" ")

    # ----------------------------------------------------------------------
    typer.echo("--")



"""Utility to run 'cmd' from python as a subprocess"""

import subprocess
import re
from pathlib import Path
from typing import Optional, Union, List


def escape_ansi(line: Optional[str]) -> Optional[str]:
    """Removes ansi escape characters that come back from the terminal"""
    if line is None:
        return None

    ansi_regex = (
        r"\x1b("
        r"(\[\??\d+[hl])|"
        r"([=<>a-kzNM78])|"
        r"([\(\)][a-b0-2])|"
        r"(\[\d{0,2}[ma-dgkjqi])|"
        r"(\[\d+;\d+[hfy]?)|"
        r"(\[;?[hf])|"
        r"(#[3-68])|"
        r"([01356]n)|"
        r"(O[mlnp-z]?)|"
        r"(/Z)|"
        r"(\d+)|"
        r"(\[\?\d;\d0c)|"
        r"(\d;\dR))"
    )
    ansi_escape = re.compile(ansi_regex, flags=re.IGNORECASE)
    return ansi_escape.sub("", line)


def run_shell_cmd(
    cmd: str,
    capture_output: bool = False,
    print_captured_output: bool = False,
    cwd: Optional[Path] = None,
    timeout_seconds: Optional[int] = None,
    run_in_powershell: Optional[bool] = None,
) -> str:
    """Runs 'cmd' with following behavior, so we can
    use it in our sequential CI/CD pipeline:

    Example usage:
        import sys
        import subprocess
        cmd = "...some command..."
        try:
            run_shell_cmd(cmd)
        except subprocess.CalledProcessError as e:
            sys.exit(e.returncode)
        except Exception as e:
            typer.echo("ERROR: -----") # optional explanation
            sys.exit(1)

    Parameters:
        capture_output:
            False (default) - Returns ""
            True            - Returns captured stdout+stderr
                              Blocks! Cannot be used to start a background process

        print_captured_output:
            Only used when capture_output=True
            False (default) - Does NOT print output if all runs ok
                              If an error occurs, it will print it
            True            - Prints output while running

    Explanation:
        - Runs `cmd` as a shell command, as if run from the command line
        - Runs from the current folder
          (You can specify an alternative folder with 'cwd=...' argument)
        - Prints the combined stdout & stderr, while it is running

        - If no error occurs:
          - if capture_output=True, returns the captured combination of stdout & stderr
          - else, returns ""
        - If an error occurs:
          - Throws a subprocess.CalledProcessError as e:
            return e.returncode
    """
    shell = True
    if cwd is None:
        cwd = Path(".")

    if run_in_powershell is None:
        run_in_powershell = False

    # for certain commands on Windows
    cmd_: Union[str, List[str]] = cmd
    if run_in_powershell:
        cmd_ = ["powershell.exe", "-Command", cmd]

    if timeout_seconds is None:
        timeout_seconds = 5

    capture_stdout = ""

    if capture_output is False:
        subprocess.run(cmd_, shell=True, check=True, text=True, cwd=cwd)
    else:
        # These will block & cannot be used to start a background process from make.
        if print_captured_output is False:
            try:
                # This will capture_output and:
                # - return it if there is no error
                # - includes it in a possibly raised error
                p_1 = subprocess.run(
                    cmd_,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    shell=shell,
                    check=False,
                    text=True,
                    cwd=cwd,
                    timeout=timeout_seconds,
                )
                if p_1.returncode != 0:
                    # typer.echo(p_1.stdout)
                    raise subprocess.CalledProcessError(
                        returncode=p_1.returncode,
                        cmd=p_1.args,
                        output=escape_ansi(p_1.stdout),
                        stderr=escape_ansi(p_1.stderr),
                    )
                capture_stdout += p_1.stdout
            except subprocess.TimeoutExpired as e:
                capture_stdout = (
                    f"ERROR: Command '{e.cmd}' timed out after {e.timeout} seconds, "
                    f"with stdout: "
                )

                if e.stdout is not None:
                    lmax = 132
                    lstdout = len(e.stdout)
                    if isinstance(e.stdout, bytes):
                        capture_stdout += e.stdout.decode("utf-8")[: min(lmax, lstdout)]
                    else:
                        capture_stdout += e.stdout[: min(lmax, lstdout)]
                    if lstdout > lmax:
                        capture_stdout += "..."

                print(capture_output)
        else:
            # This prints output while running, and we capture it as well & return it
            # https://stackoverflow.com/a/28319191/5480536
            with subprocess.Popen(
                cmd_,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                shell=shell,
                bufsize=1,
                universal_newlines=True,
                cwd=cwd,
            ) as p_2:
                if p_2.stdout is not None:
                    for line in p_2.stdout:
                        print(line, end="")
                        capture_stdout += line
            if p_2.returncode != 0:
                raise subprocess.CalledProcessError(
                    returncode=p_2.returncode,
                    cmd=p_2.args,
                    output=escape_ansi(p_2.stdout),  # type: ignore
                    stderr=escape_ansi(p_2.stderr),  # type: ignore
                )
    return capture_stdout



"""Handles 'icpp install-wasi-sdk' """
import sys
import shutil
from pathlib import Path
import math
import tarfile
import requests
import enlighten  # type: ignore
import typer
from icpp.__main__ import app
from icpp import config_default


@app.command()
def install_wasi_sdk() -> None:
    """Install wasi-sdk compiler.

    Installs the wasi-sdk compiler for icpp.
    """
    if config_default.WASI_SDK_OS_NAME == "unknown":
        typer.echo("ERROR: a wasi-sdk binary is not available for your system.")
        typer.echo(f"       - Your system: {config_default.OS_SYSTEM}")
        typer.echo(
            f"       Please follow instructions at "
            f"       https://github.com/WebAssembly/wasi-sdk to build it yourself"
            f"       and install it in {config_default.ICPP_ROOT_COMPILER}"
        )
        sys.exit(1)

    # ----------------------------------------------------------------
    typer.echo("Cleanining install folder...")

    fpath = Path(str(config_default.ICPP_ROOT_COMPILER) + ".tar.gz")
    fpath.unlink(missing_ok=True)

    try:
        shutil.rmtree(config_default.ICPP_ROOT_COMPILER)
    except FileNotFoundError:
        pass
    except OSError as e:
        typer.echo(f"Warning: {e.strerror}")

    try:
        shutil.rmtree(config_default.ICPP_ROOT_UNTAR_DIR)
    except FileNotFoundError:
        pass
    except OSError as e:
        typer.echo(f"Warning: {e.strerror}")
    # ----------------------------------------------------------------

    try:
        ################################################################################
        # Download the tar.gz file
        use_progress_bar = True

        with requests.Session() as s:
            r = s.get(config_default.WASI_SDK_URL, stream=use_progress_bar)
            if r.status_code != 200:
                typer.echo(
                    f"Tried to fetch wasi-sdk from GitHub, but response status code is "
                    f"{r.status_code}. Please retry later..."
                )
                typer.echo(f"url = {config_default.WASI_SDK_URL}")
                sys.exit(1)

            typer.echo(f"Downloading wasi-sdk: {config_default.WASI_SDK_URL}")
            typer.echo(f"Saving tar ball as  : {fpath}")

            config_default.ICPP_ROOT.mkdir(parents=True, exist_ok=True)
            if not use_progress_bar:
                with open(fpath, "wb") as f:
                    f.write(r.content)
            else:
                # https://stackoverflow.com/a/63832993/5480536
                progress_bar_manager = enlighten.get_manager()
                dlen = int(r.headers.get("Content-Length", "0")) or None

                with progress_bar_manager.counter(
                    color="cyan",
                    total=dlen and math.ceil(dlen / 2**20),
                    unit="MiB",
                    leave=False,
                ) as ctr, open(fpath, "wb", buffering=2**24) as f:
                    for chunk in r.iter_content(chunk_size=2**20):
                        # print(chunk[-16:].hex().upper())
                        f.write(chunk)
                        ctr.update()

                progress_bar_manager.stop()

        ################################################################################
        # Unzip the tar.gz file
        typer.echo(f"Untarring {fpath}")
        with tarfile.open(fpath, "r") as tar:
            tar.extractall(path=config_default.ICPP_ROOT)

        if config_default.ICPP_ROOT_UNTAR_DIR != config_default.ICPP_ROOT_COMPILER:
            typer.echo(
                f"Renaming {config_default.ICPP_ROOT_UNTAR_DIR} to "
                f"{config_default.ICPP_ROOT_COMPILER}"
            )
            try:
                # Convert Path to str, for compatibility with Python 3.8
                shutil.move(
                    str(config_default.ICPP_ROOT_UNTAR_DIR),
                    str(config_default.ICPP_ROOT_COMPILER),
                )
            except shutil.Error as e:
                typer.echo(f"Error: {e}")
                sys.exit(1)

        typer.echo("Cleaning up")
        fpath.unlink(missing_ok=True)

        typer.echo(
            f"Successfully installed wasi-sdk in {config_default.ICPP_ROOT_COMPILER}"
        )
        try:
            typer.echo("💯 🎉 🏁")
        except UnicodeEncodeError:
            typer.echo(" ")
        typer.echo("--")

    except Exception as e:  # pylint: disable=broad-except
        typer.echo(
            f"Tried to fetch wasi-sdk from GitHub at url={config_default.WASI_SDK_URL} "
            f"and install in {config_default.ICPP_ROOT_COMPILER}, but received this "
            f"Error: {e}"
        )
        sys.exit(1)


def is_wasi_sdk_installed() -> bool:
    """Returns True if the wasi-sdk is installed."""
    if config_default.ICPP_ROOT_COMPILER.exists():
        return True

    return False



"""
Handles licensing of icpp-pro

********************************************************************************
* IT IS NOT ALLOWED TO MODIFY THIS FILE OR ANY FILE THAT CALLS THESE FUNCTIONS *
********************************************************************************

"""
import sys
from typing import Optional
import typer


def exit_if_not_pro(capability: Optional[str] = None) -> None:
    """Exit if not running a licensed icpp-pro."""
    if capability is None:
        msg = "this command"
    else:
        msg = f"'{capability}'"
    if not is_pro():
        typer.echo("--")
        typer.echo(f"ERROR: {msg} requires icpp-pro")
        typer.echo(" ")
        typer.echo("For licensing, contact us at icpp@icpp.world")
        sys.exit(1)


def is_pro() -> bool:
    """Returns true if running a licensed version of icpp-pro."""
    return True



"""Test canister APIs

   First deploy the canister, then run:

   $ pytest --network=[local/ic] test_apis.py

"""
# pylint: disable=missing-function-docstring, unused-import, wildcard-import, unused-wildcard-import, line-too-long

from pathlib import Path
from typing import Dict
import pytest
from icpp.smoketest import call_canister_api, dict_to_candid_text

# Path to the dfx.json file
DFX_JSON_PATH = Path(__file__).parent / "../dfx.json"

# Canister in the dfx.json file we want to test
CANISTER_NAME = "greet"


def test__greet_0(network: str) -> None:
    response = call_canister_api(
        dfx_json_path=DFX_JSON_PATH,
        canister_name=CANISTER_NAME,
        canister_method="greet_0",
        canister_argument="()",
        network=network,
    )
    expected_response = '("hello!")'
    assert response == expected_response


# Run this test with anonymous identity
def test__greet_0_auth_err(identity_anonymous: Dict[str, str], network: str) -> None:
    # double check the identity_anonymous fixture worked
    assert identity_anonymous["identity"] == "anonymous"
    assert identity_anonymous["principal"] == "2vxsx-fae"

    response = call_canister_api(
        dfx_json_path=DFX_JSON_PATH,
        canister_name=CANISTER_NAME,
        canister_method="greet_0_auth",
        canister_argument="()",
        network=network,
    )
    expected_response = "(variant { err = 401 : nat16 })"
    assert response == expected_response


# Run this test with a logged in default identity
def test__greet_0_auth_ok(identity_default: Dict[str, str], network: str) -> None:
    # double check the identity_anonymous fixture worked
    assert identity_default["identity"] == "default"

    response = call_canister_api(
        dfx_json_path=DFX_JSON_PATH,
        canister_name=CANISTER_NAME,
        canister_method="greet_0_auth",
        canister_argument="()",
        network=network,
    )
    principal = identity_default["principal"]
    expected_response = f'(variant {{ ok = "Hello {principal}" }})'
    assert response == expected_response


def test__greet_1(network: str) -> None:
    response = call_canister_api(
        dfx_json_path=DFX_JSON_PATH,
        canister_name=CANISTER_NAME,
        canister_method="greet_1",
        canister_argument="()",
        network=network,
    )
    expected_response = "(2_023 : int)"
    assert response == expected_response


def test__greet_2(network: str, principal: str) -> None:
    response = call_canister_api(
        dfx_json_path=DFX_JSON_PATH,
        canister_name=CANISTER_NAME,
        canister_method="greet_2",
        canister_argument='("C++ Developer")',
        network=network,
    )
    expected_response = f'("hello C++ Developer!\\nYour principal is: {principal}")'
    assert response == expected_response


def test__greet_3(network: str) -> None:
    response = call_canister_api(
        dfx_json_path=DFX_JSON_PATH,
        canister_name=CANISTER_NAME,
        canister_method="greet_3",
        canister_argument='(record { "icpp version" = 1 : int; OS = "Linux" : text })',
        network=network,
    )
    expected_response = '(record { "icpp Release Details" = "Version = 1 & Operating System = Linux"; "release year" = 2_023 : int;})'
    assert response == expected_response


def test__greet_4(network: str) -> None:
    response = call_canister_api(
        dfx_json_path=DFX_JSON_PATH,
        canister_name=CANISTER_NAME,
        canister_method="greet_4",
        canister_argument="(record { 6 = 42 : int; 9 = 43 : int }, record { 7 = 44 : int; 10 = 45 : int })",
        network=network,
    )
    expected_response = (
        '("Hello!", "Your secret numbers are:", 42 : int, 43 : int, 44 : int, 45 : int)'
    )
    assert response == expected_response


def test__greet_json(network: str, principal: str) -> None:
    d = {"name": "AJB"}
    text_in = dict_to_candid_text(d)
    response = call_canister_api(
        dfx_json_path=DFX_JSON_PATH,
        canister_name=CANISTER_NAME,
        canister_method="greet_json",
        canister_argument=text_in,
        network=network,
    )
    expected_response = '("{\\"greet\\":\\"Hello AJB!\\"}")'
    assert response == expected_response
